/* 
  Program: manets_simulation.c
  Author : Vaidehi 
  Notes  :
    1. Sender knowns the number of neighboring nodes available to them.
    2. Routing protocol ditermines the disjoint paths between 
       the sender and the receiver.
    3. If one of the nodes in the disjoint path dies then the 
       disjoint path is not available.
    4. The number of splits are always lesser than the number of 
       available neighbors to the sender.
    5. Minimum Requirements:SENDER_MSG_SPLIT_COUNT = 1
                            NEIGHBORING_NODES = 2  
                            LEVELS = 1

*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <mqueue.h>
#include <gmp.h>
#include <semaphore.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "manets_simulation.h"

int main(int argc, char *argv[])
{
  int c;
  char *eval = NULL;

  while(( c = getopt(argc, argv, "rme:k"))!=-1)
    switch(c){
      case 'r':
        printf("Invoke the sender, router, destination\n");
        start_nodes();
        break;
      case 'm':
        parent_to_sender_msg();
        break;
      case 'e':
        eval = optarg;

        if (strcmp(eval, "da")  == 0){
          set_encryption_scheme(DA);
          set_df_add_mul_ph_keys();
        } else if (strcmp(eval, "dn") == 0) {
          set_encryption_scheme(DN);
          set_df_new_ph_keys();
        } else if (strcmp(eval, "df") == 0) {
          set_encryption_scheme(DF);
          set_df_field_mmh_ph_keys();
        } else if(strcmp(eval, "mmh") ==0){
          set_encryption_scheme(MMH);
          set_df_field_mmh_ph_keys();
        }else{
          fprintf (stderr, "Invalid arguments to option e                                                           Valid arguments: da, dn, df, mmh \n");
        }
        break;
      case '?':
        if (isprint (optopt))
          fprintf (stderr, "Use option -r:start up routers, sender & receiver                                     -m: send message.\n", optopt);
          else
          fprintf (stderr, 
                        "Unknown option character `\\x%x'.\n", 
                        optopt);
        return 1;
      default:
        abort();
    }//switch

  return 0;
    
}


void start_nodes(){

  int i;

  /*Create and initialize the route table*/
  create_shared_memory();
  init_route_table();

  /*Create and initialize the encryption keys*/
  init_enc_keys();

  /* Start the sender */
  sender();


  /* Start the routers */
  for(j=0; j<MAX_NODES-2; j++)
  router();


  /* Start the receiver */
  receiver();


  signal(SIGINT, parenthandler);

  for(i=0;i<MAX_NODES;i++)
    printf("pid's :%d\n", childpid[i]);

  for(i=0;i<MAX_NODES;i++)
      //waitpid(-1, NULL, 0|WUNTRACED);
    waitpid(childpid[i], NULL, 0|WUNTRACED);

  exit(0);
}

void parent_to_sender_msg(){
  struct mq_attr get_set_attr;
  struct mq_attr get_attr;
  char sender_mqueue[MAX_NODES][100];
  int nwrite;
  size_t count;
  mpz_t str, test;
  mpz_init(str);
  mpz_init(test);
  size_t size = sizeof(char);
  size_t numb = 8*size - 0;
  char mpz_buff[800]={0};
  int *pInt = (int *)mpz_buff;
  int impcount=0;
  int cmp=0;
  unsigned int time=0;
  int group, number;

  /*setting the atributes of mq_attr*/
  get_attr.mq_msgsize = 800;
  get_attr.mq_maxmsg = 5;

  /* Get the sender message queue*/
  get_sender_mqueue(sender_mqueue);

  parent_write_fd = mq_open(sender_mqueue[0], O_WRONLY);
  if(parent_write_fd == -1)
  {
    errfunction("[parent_to_sender_msg]:cannot open message queue.");
  }
    
  /* Reading message from stdinput and sending data to sender msgqueue*/
  enc_scheme encryption_scheme = get_encryption_scheme();

  printf("manets_simulation> ");

  while(mpz_inp_str(str, NULL, 10)!=0){


    if((mpz_cmp_ui(str, 11)==0 || mpz_cmp_ui(str, 12)==0 || 
        mpz_cmp_ui(str, 13)==0 || mpz_cmp_ui(str, 14)==0) && 
       (encryption_scheme == DA))
    {
      set_df_mul_ph_rd_keys(str);
    }
    else if((mpz_cmp_ui(str, 11)==0 || mpz_cmp_ui(str, 12)==0 || 
             mpz_cmp_ui(str, 13)==0 || mpz_cmp_ui(str, 14)==0 ||
             mpz_cmp_ui(str, 15)==0)&& 
            (encryption_scheme == DN))
    {
      set_df_new_ph_rp_p_d_keys(str);
    }
    else if((mpz_cmp_ui(str, 11)==0 )&& (encryption_scheme == DF))
    {
      set_df_field_mmh_ph_p_r_keys(str);
    }
    else if((mpz_cmp_ui(str, 11)==0 )&& (encryption_scheme == MMH))
    {
      set_df_field_mmh_ph_p_r_keys(str);
    }
    else if(mpz_cmp_ui(str, 2)==0)
    {
      printf("manets_simulation> Enter the sleep time in sec:");
      scanf("%d", &time);
      sleep(time);
    }
    else
    {
      count = (mpz_sizeinbase (str, 2) + numb-1) / numb;
      pInt[0]=count;
      mpz_export(mpz_buff+4, NULL, 1, 1, 0, 0, str);
      mpz_import(test, count, 1, 1, 0, 0, mpz_buff+4);
      nwrite= mq_send(parent_write_fd, mpz_buff, sizeof(mpz_buff), 0);
      if(nwrite == -1)
        errfunction("[parent_to_sender_msg]:cannot send the message");
      }  
  printf("manets_simulation> ");
  }


  mq_close(parent_write_fd);
  mpz_clear(str);
  mpz_clear(test);

}



/* Signal handler of parent */
void parenthandler(int signum){

  printf("The parent signal handler is invoked\n");
  int i;
  for(i=0;i<MAX_NODES;i++){
    kill(childpid[i], SIGINT);
      //waitpid(childpid[i], NULL, 0|WUNTRACED);
    //printf("%d\n", childpid[i]);
  }  
  for(i=0;i<MAX_NODES;i++){
    waitpid(childpid[i], NULL, 0|WUNTRACED);
  }
  shm_unlink(ROUTER_TABLE);
  sem_unlink(ROUTER_TABLE_LOCK);
  shm_unlink(ENC_KEYS_TABLE);
  sem_unlink(ENC_KEYS_TABLE_LOCK);
  printf("The parent sent signal to kids and closed fd \n");
  exit(0);  
}

/* Signal handler of sender */
void senderhandler(int signum){
  int i=0;

  printf("[Sender] handle is invoked\n");
  mq_close(sender_read_fd);
  mq_close(sender_write_fd);
  mq_unlink(sender_mqueue);
  printf("[Sender] handle closed all discriptors\n");
  exit(0);

}

/* Signal handler of router */
void routerhandler(int signum){

  printf("[Router] handle is invoked\n");
  mq_close(router_write_fd);
  mq_close(router_read_fd);
  mq_unlink(router_mqueue);
  exit(0);

}

/* Signal handler of receiver */
void receiverhandler(int signum){

  printf("[Receiver] handle is invoked\n");
  mq_close(receiver_read_fd);  
  mq_unlink(receiver_mqueue);
  exit(0);
}


void sender(){
  int pid;
  
  pid = fork();
  if ( pid == 0 )
  {
    signal(SIGINT, senderhandler);
    // Set the path for encryption log file
    set_manets_log((char *)ENC_LOG_FILE); 
    unlink(manets_log);
    printf(" The path is: %s",manets_log);

    // child process
    struct mq_attr get_set_attr; 
    struct mq_attr get_attr; 
    get_attr.mq_msgsize = 8000;
    get_attr.mq_maxmsg = 8;
    int attr;
    ssize_t nread;
    int nwrite;

    char nodes_mqueues[MAX_NODES][100];
    int neighbouring_groups[100];
    int neighbouring_count;
    int group_id, group_nodes_id;

    mpz_t sender_msg_split_array[50];

    for(j=0;j<50;j++){
      mpz_init(sender_msg_split_array[j]);
    }

    char senderbuff[8000]={0}; 
    int message_id; 
    int message_split_id;
    int message_split_id_sum=0;
    int base=10;
    size_t size = sizeof(char);
    size_t numb = 8*size - 0;
    long seed;
    int i, j;
    int k;
    int cmp;
    int num;
    int counter;
    char sender_msg_split_buff[255][8000];
    mpz_t cipher[255][255]; 
    mpz_t msg;
    int count;

    mpz_t temp, sum;
    enc_scheme encryption_scheme;

    mpz_symmetric_key mpz_keys_ptr;

    int intarr[255];
    int isum;
    mpz_t quotient, remainder;

    /* Encryption Variables*/
    int encryption_split_number;

    printf("[Sender] Enter forked  process ... \n");
    /* Sender sets the route table*/
    add_node(getpid());

    /*Sender creating the message queue to read the message to be sent 
       to the  receiver*/
    sprintf(sender_mqueue, "/route.%d", getpid());
    sender_read_fd=mq_open(sender_mqueue, O_RDWR|O_CREAT, 0666, &get_attr);
    if(sender_read_fd == -1){
      errfunction("sender");
    }
    attr = mq_getattr(sender_read_fd, &get_set_attr);
    printf("[Sender] max size of msg %d\n", get_set_attr.mq_msgsize);
    printf("[Sender] opened the message queue:%s\n", sender_mqueue);
    

    /* Sender reads the mesage to send to the router*/
    
    message_id=0;

    for(i=0;i<255;i++){
      for(j=0;j<255;j++){
        mpz_init(cipher[i][j]);
      }
    }
    mpz_init(msg);
    mpz_init(quotient);
    mpz_init(remainder);

    while((nread = mq_receive(sender_read_fd, senderbuff, 
                              get_set_attr.mq_msgsize, NULL))>0){
      printf("[Sender] Reads the msg sent by user\n");
      if(nread == -1){
        errfunction("sender can't read user msg");
      }

      /* Converting the char buffer "senderbuff" to an mpz_t 
         variable*/
      int *pInt = (int *)senderbuff;
      count = pInt[0];

      printf("Size of the message:%d\n", count);
      printf("[Sender] Importing\n");
      mpz_import(msg, count, 1, sizeof(senderbuff[0]), 0, 0, senderbuff+4);
      printf("[Sender] Imported\n");



       seed = get_seed();
     
       /* Randomly generate the encryption split message values */
       srand(seed);
       mpz_sub_ui(msg, msg, 32767*SENDER_MSG_SPLIT_COUNT);
       cmp =0;
       int k = 0;
       while(cmp==0){
         isum=0;
         for(j=0;j<SENDER_MSG_SPLIT_COUNT-1;j++){
           intarr[j]=1+(int) (32767.0*rand()/(RAND_MAX+1.0));
           isum = isum+ intarr[j];
         }
         if(isum > 32767*SENDER_MSG_SPLIT_COUNT)
         {
           cmp=0;
         }
           else
         {
           cmp=1;
         }
         printf("Entered gen_random %d , %d\n", k, isum);
         k++;
       }
     
       intarr[j] = 32767*SENDER_MSG_SPLIT_COUNT - isum;
       mpz_tdiv_qr_ui(quotient, remainder, msg, SENDER_MSG_SPLIT_COUNT);
       for(j=0;j<SENDER_MSG_SPLIT_COUNT;j++){
         mpz_set(sender_msg_split_array[j], quotient);
         mpz_add_ui(sender_msg_split_array[j], sender_msg_split_array[j], 
                    intarr[j]);
       }
     
       mpz_add(sender_msg_split_array[j-1], sender_msg_split_array[j-1], 
               remainder);

      /* Randomly generate the split message values and put it into
         sender_msg_split_array*/
      mpz_init_set_str(temp, "0", base);
      mpz_init_set_str(sum, "0", base);

      printf(" The split messages");
      for(j=0;j<SENDER_MSG_SPLIT_COUNT;j++){
        gmp_printf("%Zd\n", sender_msg_split_array[j]);
      }

      /* Encrypt the split messages */
      printf(" Enter the get various encryption keys\n");

      get_various_encryption_keys(&mpz_keys_ptr);

      printf(" Exit the get various encryption keys\n");
      message_split_id_sum = (SENDER_MSG_SPLIT_COUNT *
                             (SENDER_MSG_SPLIT_COUNT+1))/2;
      for(i=0;i<SENDER_MSG_SPLIT_COUNT;i++){
        //mpz_set(splitmsg, sender_msg_split_array[i]);
        encrypt(sender_msg_split_array[i], sender_msg_split_buff[i], 
                message_id, i+1, message_split_id_sum, &mpz_keys_ptr);

      }

      /* Get the sender's neighboring nodes */
      get_next_node(getpid(), nodes_mqueues, neighbouring_groups, 
                    &neighbouring_count);

      num=0;
      counter =0;
      printf(" Message split_id sum:%d\n", message_split_id_sum);

      while(num < neighbouring_count){
        if(counter==SENDER_MSG_SPLIT_COUNT){
          counter =0;
        }
          
        group_id =  neighbouring_groups[num];
        group_nodes_id = group_id;
        printf("**********************************************************\n");
        printf("The group id:%d at position num:%d have msg counter:%d\n", 
               neighbouring_groups[num], num, counter);  
        printf("**********************************************************\n");

        while(group_nodes_id == group_id && num < neighbouring_count){
         
          sender_write_fd = mq_open(nodes_mqueues[num], O_WRONLY);
          nwrite= mq_send(sender_write_fd, sender_msg_split_buff[counter], 
                          sizeof(sender_msg_split_buff[counter]), 0);
          printf("[Sender] writes the msg to the router msgqueue\n");
          if(nwrite == -1){
            errfunction("sender cannot write msg to router");
          }

          mq_close(sender_write_fd);

          num++;
          if(num < neighbouring_count){
            group_nodes_id = neighbouring_groups[num];
          }

        }
        counter++;
      }
      message_id++;
      mpz_clear(mpz_keys_ptr.r);
      mpz_clear(mpz_keys_ptr.md);
      mpz_clear(mpz_keys_ptr.m);
      mpz_clear(mpz_keys_ptr.p);
      mpz_clear(mpz_keys_ptr.q);
      mpz_clear(mpz_keys_ptr.rp);
      mpz_clear(mpz_keys_ptr.rq);
      mpz_clear(mpz_keys_ptr.mpq);

    }

    mpz_clear(msg);
    mpz_clear(quotient);
    mpz_clear(remainder);
    for(i=0;i<255;i++){
      for(j=0;j<255;j++){
        mpz_clear(cipher[i][j]);
      }
    }

    for(j=0;j<50;j++){
      mpz_clear(sender_msg_split_array[j]);
    }
    mq_close(sender_read_fd);
    mq_close(sender_write_fd);
    mq_unlink(sender_mqueue);
    mpz_clear(temp);
    mpz_clear(sum);

    exit(0);

  } 
  else if ( pid >0 ) // PID
  {
    printf("[Parent(s) After forking Sender Child process ... \n");
    printf("[Parent(s)Sender Child pid : %d\n", pid);
    childpid[0] = pid;
  }
  else
  {
    printf("[Parent(s)]fork error\n");
    exit(1);
  }
}


void receiver(){
  int pid;

  pid = fork();

  if(pid==0){
   signal(SIGINT, receiverhandler);
   set_manets_log((char *)DEC_LOG_FILE);
   unlink(manets_log);
   printf(" The path is: %s",manets_log);

   // child process
   printf("[Receiver]Entering the receiver process\n");
   struct mq_attr get_attr, get_set_attr; 
   get_attr.mq_msgsize = 8000;
   get_attr.mq_maxmsg = 1;
   int n;
   ssize_t nread;
   char receivebuff[8000]={0};

   int message_id=0;
   int message_split_id=0;
   int message_split_id_sum=0;
   int pmessage_split_id_sum=0;
   int split_id_sum=0;
   int message_split_id_array[500]={0};
   int counter=0, enter=0;
   int previous_counter=0;
   int i, j;
   int current_message_id=0;
   enc_scheme encryption_scheme;
   mpz_symmetric_key mpz_keys_ptr;


   /* Decryption*/
   int encryption_split_number ;
   mpz_t encryption_msg_split_array[255][255];
   mpz_t plaintext;
   mpz_init(plaintext);
   for(i=0;i<255;i++){
     for(j=0;j<255;j++){
       mpz_init(encryption_msg_split_array[i][j]);
     }  
   }


   /*Receiver sets the route table*/
   add_node(getpid());


   /* The receiever creates the message queue to read the message from the
    * routers*/
   sprintf(receiver_mqueue, "/route.%d", getpid());
   receiver_read_fd = mq_open(receiver_mqueue, O_RDWR|O_CREAT, 0666, &get_attr);
   if(receiver_read_fd == -1){
     errfunction("receiver");
   }
   mq_getattr(receiver_read_fd, &get_set_attr);
   printf("[Receiver] The msg size: %d\n", get_set_attr.mq_msgsize);
   printf("[Receiver] Enter receiver Child process and opened message queue:%s\n", 
          receiver_mqueue);
   
   

   counter =0; 
    /* The receiever reads the message sent by the sender*/
   while(
   (nread = mq_receive(receiver_read_fd, receivebuff, get_set_attr.mq_msgsize, 
                       NULL))>0
   ){

     get_various_encryption_keys(&mpz_keys_ptr);
     encryption_scheme = mpz_keys_ptr.Ph;
     encryption_split_number = mpz_keys_ptr.d;

     printf("[Receiver] Reads the data from the router\n");
     if(nread == -1){
       errfunction("receiver");
     }
       
     int *int_buff = (int *)receivebuff;
     short int *short_int_buff = (short int *)receivebuff;

     message_id = receivebuff[0];
     message_split_id = receivebuff[1];
     message_split_id_sum = short_int_buff[1];

     /* Condition executed only for the first message*/
     if(counter ==0){
       current_message_id = message_id;
       message_split_id_array[counter] = message_split_id;
       buff_to_mpz_array(receivebuff, 
                         encryption_msg_split_array[message_split_id-1], 
                         encryption_split_number, encryption_scheme);
       split_id_sum = split_id_sum+message_split_id;
       pmessage_split_id_sum = message_split_id_sum;
       enter=0;
       counter++;
     }

     /* Condition executed when the message is yet to be receoverd  */
     if(current_message_id == message_id && split_id_sum!=message_split_id_sum){
       pmessage_split_id_sum = message_split_id_sum;
       enter=0;
       /* Check the message_split_id_array to see if the 
        message_split_id already exists*/
       for(i=0;i<counter;i++){
         if(message_split_id_array[i] == message_split_id){
           i++;
           break;
         }  
       }

       /* Condition executed when the split message is new*/

       if(message_split_id_array[i-1] != message_split_id){
         current_message_id = message_id;
         message_split_id_array[counter]=message_split_id;
         split_id_sum = split_id_sum+message_split_id;

         buff_to_mpz_array(receivebuff, 
                           encryption_msg_split_array[message_split_id-1], 
                           encryption_split_number, encryption_scheme);
         counter++;
       }

     }

     /* Condition executed when the message is recovered. The recoverd message
      * is then decrypted*/
     if(current_message_id == message_id && split_id_sum == message_split_id_sum 
        && enter==0){
       pmessage_split_id_sum = message_split_id_sum;
       printf(" The Message is:");
       decrypt(encryption_msg_split_array, plaintext, counter, &mpz_keys_ptr);
       gmp_printf("The message is %Zd\n", plaintext);
       enter=1;
     }

     /* Condition executed for the new message when the previous message is
      * recoverd */
     if(current_message_id != message_id && 
        split_id_sum==pmessage_split_id_sum )
     {
       current_message_id = message_id;
       message_split_id_array[0]=message_split_id;
       counter = 1;
       enter=0;
       split_id_sum=0;
       split_id_sum = split_id_sum+message_split_id;
       buff_to_mpz_array(receivebuff, 
                         encryption_msg_split_array[message_split_id-1], 
                         encryption_split_number, encryption_scheme);
       //decrypt(encryption_msg_split_array, plaintext, counter, &mpz_keys_ptr);
     }

     /* Condition executed for the new message when the previous message is
      * not recovered */
     if(current_message_id != message_id && 
        split_id_sum!=pmessage_split_id_sum)
     {
       printf(" The Previous message is not recovered\n");
       current_message_id = message_id;
       message_split_id_array[0]=message_split_id;
       counter = 1;
       split_id_sum=0;
       split_id_sum = split_id_sum+message_split_id;
       buff_to_mpz_array(receivebuff, 
                         encryption_msg_split_array[message_split_id-1], 
                         encryption_split_number, encryption_scheme);
     }
        
     mpz_clear(mpz_keys_ptr.r);
     mpz_clear(mpz_keys_ptr.md);
     mpz_clear(mpz_keys_ptr.m);
     mpz_clear(mpz_keys_ptr.p);
     mpz_clear(mpz_keys_ptr.q);
     mpz_clear(mpz_keys_ptr.rp);
     mpz_clear(mpz_keys_ptr.rq);
     mpz_clear(mpz_keys_ptr.mpq);
   }//while

   mq_close(receiver_read_fd);  
   mq_unlink(receiver_mqueue);

   mpz_clear(plaintext);
   for(i=0;i<255;i++){
     for(j=0;j<255;j++){
       mpz_clear(encryption_msg_split_array[i][j]);
     }  
   }

   exit(0);
  }//pid==0
  else if(pid>0)
  {
   printf("[Parent(r)] Receiver Child pid : %d\n", pid);
   childpid[MAX_NODES-1]=pid;
  }
  else
  {
     printf("[Parent(r)] fork error\n");
     exit(1);
  }
}


void router(){
  int pid;

  pid = fork();
  
  if(pid == 0){
   signal(SIGINT, routerhandler);
   signal(SIGALRM, routerhandler);

   char node_mqueues[MAX_NODES][100];
   int neighbouring_groups[100];
   int neighbouring_count;
   int count=0;

   struct mq_attr get_attr; 
   get_attr.mq_msgsize = 8000;
   get_attr.mq_maxmsg = 1;
   ssize_t nread;
   int nwrite;
   char readbuff[8000]={0};


   /* variables used to generate the random numbers with exponential
    * distribution*/
   long seed;
   const gsl_rng_type *T;
   gsl_rng *r;
   double mu =EXP_DIST_MEAN;
   unsigned long rand_exp;
   unsigned int sec;

   /* Set the route table by the router*/
   add_node(getpid());

   sprintf(router_mqueue, "/route.%d", getpid());

   /* Mqueue created by router to receive message sent by the sender or other
    * routers*/
   router_read_fd = mq_open(router_mqueue, O_RDWR|O_CREAT, 0666, &get_attr);
   printf("[Router] max size of msg %d\n", get_attr.mq_msgsize);
   if(router_read_fd == -1){
     errfunction("router");
   }
   printf("[Router: %d] The router message queue %s created\n", 
           getpid(), router_mqueue);  



   seed = get_seed();

   T= gsl_rng_default;
   r= gsl_rng_alloc(T);
   gsl_rng_set(r, seed);
   rand_exp = gsl_ran_exponential(r, mu);
   gsl_rng_free(r);

   /* Reading the message sent by the sender to the router queue*/
   while((nread = mq_receive(router_read_fd, readbuff, get_attr.mq_msgsize, 
                             NULL))>0)
   {
     printf("[Router] Reads the msg sent by sender\n");

     if(nread == -1){
       errfunction("router");
     }
     printf("[Router : %d] Read the message send from sender by the router..\n",
            getpid());
     
     if(rand_exp == 0){
       rand_exp = 1;
     }
#ifndef NO_ALARM
     sec = (unsigned int)rand_exp;  
     alarm(rand_exp);
#endif

     printf("Alarm time:%d \n", rand_exp);


     /* Reads the route table to get the neighbouring nodes' message queues*/
     get_next_node(getpid(), node_mqueues, neighbouring_groups, 
                   &neighbouring_count);
   
     /* Open neighbouring nodes' message queue and send the message to the 
     nodes
     */
     count=0;
     while(count<neighbouring_count){

       router_write_fd = mq_open(node_mqueues[count], O_WRONLY);
       if(router_write_fd == -1){
         errfunction("[router: opening neighbouring nodes' message queue]");
       }
       printf("[Router: %d]  Opened neighbouting node message queue....\n", 
              getpid());
       nwrite= mq_send(router_write_fd, readbuff, sizeof(readbuff), 0);
       printf("[Router] writes the msg to the neighbouring node\n");
       if(nwrite == -1){
         errfunction("router");
       }
       count++;

       mq_close(router_write_fd);
     }
   }
   mq_close(router_read_fd);
   mq_unlink(router_mqueue);

   exit(0);
   
  }
  else if(pid >0){
   //kill(pid, SIGSTOP);
   printf("[Parent(ro) Router pid %d\n", pid);
   childpid[j+1]=pid;
  }
  else if(pid<0){
   printf("[Parent(ro ) error\n");
  }
}


void errfunction(char error_loc[]){
   int errsv = errno;
   printf("[%s] Error : %s\n", error_loc, strerror(errsv));
   //exit(1);
}


void create_shared_memory(){
  /* Create the shared memory */

  int sk_d, rt_d;

  /* Remove the already existed shared memory and create it*/
  shm_unlink(ENC_KEYS_TABLE);
  sk_d = shm_open(ENC_KEYS_TABLE, O_CREAT|O_RDWR, 0666);
  ftruncate(sk_d, sizeof(symmetric_key));
  close(sk_d);

  sem_unlink(ENC_KEYS_TABLE_LOCK);
  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, O_CREAT, 0666, 1);


  /* Remove the already existed shared memory and create it*/

  shm_unlink(ROUTER_TABLE);
  rt_d = shm_open(ROUTER_TABLE, O_CREAT|O_RDWR, 0666);
  ftruncate(rt_d, sizeof(route_table));
  close(rt_d);

  sem_unlink(ROUTER_TABLE_LOCK);
  node_mutex = sem_open(ROUTER_TABLE_LOCK, O_CREAT, 0666, 1);

}


void init_enc_keys(){

  /* Open the shared memory and set up the keys in shared memory*/

  symmetric_key *sk_ptr;
  /*Function that opens the shared memory*/ 
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    /* The keys for df_addive_multiplicative Ph*/
    mpz_t r, md, m; 

    /* Set some defaults */
    mpz_init_set_str(r, "3", 10);
    mpz_init_set_str(md, "7", 10);
    mpz_init_set_str(m, "28", 10);

    /* The keys for df_new Ph*/
    mpz_t rp, rq, p, q, mpq; 

    /* Set some defaults */
    mpz_init_set_str(p, "13", 10);
    mpz_init_set_str(q, "11", 10);
    mpz_init_set_str(rp, "2", 10);
    mpz_init_set_str(rq, "3", 10);
    mpz_init(mpq);
    mpz_mul(mpq, p, q);

    long size = sizeof(char);
    long numb = 8*size - 0;

    sk_ptr->r_count = (mpz_sizeinbase (r, 2) + numb-1) / numb;
    sk_ptr->md_count = (mpz_sizeinbase (md, 2) + numb-1) / numb;
    sk_ptr->m_count = (mpz_sizeinbase (m, 2) + numb-1) / numb;

    sk_ptr->p_count = (mpz_sizeinbase (p, 2) + numb-1) / numb;
    sk_ptr->q_count = (mpz_sizeinbase (q, 2) + numb-1) / numb;
    sk_ptr->rp_count = (mpz_sizeinbase (rp, 2) + numb-1) / numb;
    sk_ptr->rq_count = (mpz_sizeinbase (rq, 2) + numb-1) / numb;
    sk_ptr->mpq_count = (mpz_sizeinbase (mpq, 2) + numb-1) / numb;

    mpz_export(sk_ptr->r_buff, NULL, 1, 1, 0, 0, r);
    mpz_export(sk_ptr->md_buff, NULL, 1, 1, 0, 0, md);
    mpz_export(sk_ptr->m_buff, NULL, 1, 1, 0, 0, m);

    mpz_export(sk_ptr->p_buff, NULL, 1, 1, 0, 0, p);
    mpz_export(sk_ptr->q_buff, NULL, 1, 1, 0, 0, q);
    mpz_export(sk_ptr->rp_buff, NULL, 1, 1, 0, 0, rp);
    mpz_export(sk_ptr->rq_buff, NULL, 1, 1, 0, 0, rq);
    mpz_export(sk_ptr->mpq_buff, NULL, 1, 1, 0, 0, mpq);

    sk_ptr->d = 3;

  sem_post(enc_mutex);
  sem_close(enc_mutex);
  mpz_clear(r);
  mpz_clear(md);
  mpz_clear(m); 

  mpz_clear(rp);
  mpz_clear(rq);
  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(mpq);

  munmap(sk_ptr, sizeof(symmetric_key));
}


void df_add_mul_Ph_encryption(mpz_t splitmsg, char sender_msg_split_buff[], 
                              int message_id, int message_split_id, 
                              int message_split_id_sum, 
                              mpz_symmetric_key *mpz_keys_ptr,
                              FILE *encrypt_time_file_handle) 
{ 

  mpz_t r, m, kr, md, product, quotient, remainder, zero;
  mpz_t encryption_msg_split_array[255];
  mpz_t cipher[255];
  int base = 10;
  long encryption_split_number;
  int i,j;
  int cmp;
  unsigned int isum=0;
  unsigned int intarr[255];
  long seed;

  size_t *countp=NULL;
  int  buffappend=0;
  int header;
  int zerocmp;
  int encryption_msg_split_size;
  size_t size = sizeof(char);
  size_t numb = 8*size - 0;

  mpz_init_set_str(zero, "0", 10); 
  mpz_init(r);
  mpz_init(md);
  mpz_init(m);

  mpz_init(kr);
  mpz_init(product);
  for(j=0;j<255;j++){
    mpz_init(encryption_msg_split_array[j]);
    mpz_init(cipher[j]);
  }
  
  mpz_init(quotient);
  mpz_init(remainder);
   

  //enc_count = enc_count+1;

  mpz_set(r, mpz_keys_ptr->r);
  mpz_set(md, mpz_keys_ptr->md);
  mpz_set(m, mpz_keys_ptr->m);
  encryption_split_number =mpz_keys_ptr->d;

  seed = get_seed();

  gmp_printf(" The mesage to be split:%Zd\n", splitmsg);

  /* Randomly generate the encryption split message values */
  srand(seed);
  mpz_sub_ui(splitmsg, splitmsg, 32767);
  cmp =0;
  while(cmp==0){
    isum=0;
    for(j=0;j<encryption_split_number-1;j++){
      intarr[j]=rand()% 32767+1;
      isum = isum+ intarr[j];
    }
    if(isum > 32767)
      cmp=0;
      else
      cmp=1;
  }

  intarr[j] = 32767 - isum;
  mpz_tdiv_qr_ui(quotient, remainder, splitmsg, encryption_split_number);
  for(j=0;j<encryption_split_number;j++){
    mpz_set(encryption_msg_split_array[j], quotient);
    mpz_add_ui(encryption_msg_split_array[j], encryption_msg_split_array[j], 
               intarr[j]);
  }

  mpz_add(encryption_msg_split_array[j-1], encryption_msg_split_array[j-1], 
          remainder);

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  /*Encryption*/
  for(j=0;j<encryption_split_number;j++){
    mpz_pow_ui(kr, r, j+1);
    mpz_mul(product, encryption_msg_split_array[j], kr);
    mpz_mod(cipher[j], product, m);
  }

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Print the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                    (end_time.tv_usec-start_time.tv_usec);

  enc_msg_split_id = enc_msg_split_id + message_split_id;
  enc_sum = enc_sum + total_usecs;  
  fprintf(encrypt_time_file_handle, "Message id: %d Message_split_id: %d Time: %d uSec Sum: %d usec",message_id, message_split_id, total_usecs, enc_sum);
  if(enc_msg_split_id == message_split_id_sum){
    enc_msg_split_id=0;
    enc_count++;
    enc_avg_sum = enc_avg_sum + enc_sum;
    enc_avg = enc_avg_sum/enc_count;
    enc_sum =0;
  }
  fprintf(encrypt_time_file_handle, " Avg: %d usec\n",enc_avg);


  int *sender_msg_split_int = (int *)sender_msg_split_buff;
  short int *sender_msg_split_short_int = (short int *)sender_msg_split_buff;
  sender_msg_split_buff[0]=message_id;
  sender_msg_split_buff[1]=message_split_id;
  sender_msg_split_short_int[1]=message_split_id_sum;

  header = (encryption_split_number * 4) + 4;
  for(i=0;i<encryption_split_number;i++){
    encryption_msg_split_size = (mpz_sizeinbase(
                                 cipher[i], 2) +
                                 numb-1) / numb;
    sender_msg_split_int[i+1] = encryption_msg_split_size;
    zerocmp = mpz_cmp(cipher[i], zero);
    if(zerocmp == 0){
      sender_msg_split_buff[header+buffappend] =0;
    }
    else{
      mpz_export(sender_msg_split_buff+header+buffappend, countp, 
                   1, 1, 0, 0, 
                   cipher[i]
                   );
    }
    buffappend = buffappend+encryption_msg_split_size;
  }

  mpz_clear (r);
  mpz_clear (m);
  mpz_clear (kr);
  //mpz_clear (sum);
  mpz_clear (product);
  mpz_clear (quotient);
  mpz_clear (remainder);
  mpz_clear (zero);
  for(j=0;j<255;j++){
    mpz_clear(encryption_msg_split_array[j]);
    mpz_clear(cipher[j]);
  }

}


void df_add_mul_Ph_decryption(mpz_t encryption_msg_split_array[][255], 
                              mpz_t plaintext, int previous_counter,
                              mpz_symmetric_key *mpz_keys_ptr,
                              FILE *decrypt_time_file_handle)
{

  int i=0, j;
  mpz_t r, m, md, r_inverse, kr, product, sum;
  mpz_t partialplain[255];
  mpz_t cipher_sum[255];
  long d;
  for(i=0;i<255;i++){
    mpz_init(partialplain[i]);
    mpz_init(cipher_sum[i]);
  }
  mpz_init(r);
  mpz_init(m);
  mpz_init(md);
  mpz_init(r_inverse);
  mpz_init(kr);
  mpz_init(product);
  mpz_init(sum);
  
  dec_count = dec_count+1;

  /* Symmetric keys */
  mpz_set(r, mpz_keys_ptr->r);
  mpz_set(md, mpz_keys_ptr->md);
  mpz_set(m, mpz_keys_ptr->m);
  d =mpz_keys_ptr->d;

  /* Computing inverse of r*/
  mpz_invert(r_inverse, r, m);
 
  /* Sum up all the incomming messages*/
  mpz_set_str(sum, "0", 10);
  for(i=0;i<d;i++){
     mpz_set_str(sum, "0", 10);
     for(j=0;j<previous_counter;j++){
       mpz_add(sum, sum, encryption_msg_split_array[j][i]);
     }
     mpz_set(cipher_sum[i], sum);
       //mpz_add(sum, sum, encryption_msg_split_array[j][i]);
  }


  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  /* Decrypt the mesages*/
  mpz_set_str(sum, "0", 10);
  for(i=0;i<d;i++){
    mpz_pow_ui(kr, r_inverse, i+1);
    mpz_mul(product, cipher_sum[i], kr);
    mpz_mod(partialplain[i], product, m);
    mpz_add(sum, sum, partialplain[i]);
    mpz_mod(plaintext, sum, md);
  }

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);

  dec_sum = dec_sum+total_usecs;
  dec_avg = dec_sum/dec_count;
  
  gmp_fprintf(decrypt_time_file_handle, "Message: %Zd ", plaintext );
  fprintf(decrypt_time_file_handle, "Time: %d uSec ", total_usecs);
  fprintf(decrypt_time_file_handle, "Sum: %d uSec ", dec_sum);
  fprintf(decrypt_time_file_handle, "Avg: %d uSec\n ", dec_avg);

  mpz_clear (r);
  mpz_clear (m);
  mpz_clear (kr);
  mpz_clear (md);
  mpz_clear (r_inverse);
  mpz_clear (sum);
  mpz_clear (product);
  for(i=0;i<255;i++){
    mpz_clear(partialplain[i]);
    mpz_clear(cipher_sum[i]);
  }
 
}


void df_new_Ph_encryption(mpz_t splitmsg, char sender_msg_split_buff[], 
                          int message_id, int message_split_id, 
                          int message_split_id_sum, mpz_symmetric_key *mpz_keys_ptr, 
                          FILE *encrypt_time_file_handle)
{

  int i, j, cmp, base=10;
  long seed;
  int encryption_split_number;
  int isum;
  int intarr[255]={0};

  mpz_t quotient, remainder, zero;
  mpz_t p, q, rp, rq, mpq;
  mpz_t sum, rp_product, rq_product, rp_exp, rq_exp;
  mpz_t encryption_msg_split_array[255];
  mpz_t cipher[255];

  size_t *countp=NULL;
  int  buffappend=0;
  int header;
  int zerocmp;
  int encryption_msg_split_size;
  gmp_randstate_t state;
  size_t size = sizeof(char);
  size_t numb = 8*size - 0;

  mpz_init(quotient);
  mpz_init(remainder); 
  mpz_init_set_str(zero, "0", 10);

  mpz_init(mpq);
  mpz_init(p);
  mpz_init(q);
  mpz_init(rp);
  mpz_init(rq);

  for(i=0;i<255;i++){
    mpz_init(encryption_msg_split_array[i]);
    mpz_init(cipher[i]);
  }

  mpz_init(sum);
  mpz_init(rp_product);
  mpz_init(rq_product);
  mpz_init(rp_exp);
  mpz_init(rq_exp);

  gmp_randinit_default(state);

  //enc_count=enc_count+1;

  mpz_set(p, mpz_keys_ptr->p);
  mpz_set(q, mpz_keys_ptr->q);
  mpz_set(rp, mpz_keys_ptr->rp);
  mpz_set(rq, mpz_keys_ptr->rq);
  mpz_set(mpq, mpz_keys_ptr->mpq);
  encryption_split_number = mpz_keys_ptr->d;

  seed = get_seed();
  gmp_randseed_ui(state, seed);

  /* Randomly generate the encryption split message values */
  srand(seed);
  mpz_sub_ui(splitmsg, splitmsg, 32767);
  cmp =0;
  while(cmp==0){
    isum=0;
    for(j=0;j<encryption_split_number-1;j++){
      intarr[j]=rand()% 32767+1;
      isum = isum+ intarr[j];
    }
    if(isum > 32767)
      cmp=0;
      else
      cmp=1;
  }
  intarr[j] = 32767 - isum;
  mpz_tdiv_qr_ui(quotient, remainder, splitmsg, encryption_split_number);

  for(j=0;j<encryption_split_number;j++){
    mpz_set(encryption_msg_split_array[j], quotient);
    mpz_add_ui(encryption_msg_split_array[j], encryption_msg_split_array[j], 
               intarr[j]);
  }
  mpz_add(encryption_msg_split_array[j-1], encryption_msg_split_array[j-1], 
          remainder);

  printf("Split values:\n");
  for(i=0;i<encryption_split_number;i++){
    gmp_printf("%Zd\n", encryption_msg_split_array[i]);
  }

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  printf("*** Encrypting with new ph***\n");

  /*Encryption*/
  for(i=0, j=1;i<(encryption_split_number*2);i=i+2, j++){
    mpz_pow_ui(rp_exp, rp, j);
    mpz_mul(rp_product, encryption_msg_split_array[j-1], rp_exp);
    mpz_mod(cipher[i], rp_product, p);
    mpz_pow_ui(rq_exp, rq, j);
    mpz_mul(rq_product, encryption_msg_split_array[j-1], rq_exp);
    mpz_mod(cipher[i+1], rq_product, q);
  }

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);

  enc_msg_split_id = enc_msg_split_id + message_split_id;
  enc_sum = enc_sum + total_usecs;  
  fprintf(encrypt_time_file_handle, "Message id: %d Message_split_id: %d Time: %d uSec Sum: %d usec",message_id, message_split_id, total_usecs, enc_sum);
  if(enc_msg_split_id == message_split_id_sum){
    enc_msg_split_id=0;
    enc_count++;
    enc_avg_sum = enc_avg_sum + enc_sum;
    enc_avg = enc_avg_sum/enc_count;
    enc_sum =0;
  }
  fprintf(encrypt_time_file_handle, " Avg: %d usec\n",enc_avg);

  printf("*** Encrypted with new ph***\n");


  int *sender_msg_split_int = (int *)sender_msg_split_buff;
  short int *sender_msg_split_short_int = (short int *)sender_msg_split_buff;
  sender_msg_split_buff[0]=message_id;
  sender_msg_split_buff[1]=message_split_id;
  sender_msg_split_short_int[1]=message_split_id_sum;

  header = ((encryption_split_number*2) * 4) + 4;
  for(i=0;i<encryption_split_number*2;i++){
    encryption_msg_split_size = (mpz_sizeinbase(
                               cipher[i], 2) +
                               numb-1) / numb;
    sender_msg_split_int[i+1]=encryption_msg_split_size;
    zerocmp = mpz_cmp(cipher[i], zero);
    if(zerocmp == 0){
      sender_msg_split_buff[header+buffappend] =0;
    }
    else{
      mpz_export(sender_msg_split_buff+header+buffappend, countp, 
                   1, 1, 0, 0, 
                   cipher[i]
                   );
    }
    buffappend = buffappend+encryption_msg_split_size;
  }

  mpz_clear(quotient);
  mpz_clear(remainder); 
  mpz_clear(zero);

  mpz_clear(mpq);
  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(rp);
  mpz_clear(rq);

  for(i=0;i<255;i++){
    mpz_clear(encryption_msg_split_array[i]);
    mpz_clear(cipher[i]);
  }

  mpz_clear(sum);
  mpz_clear(rp_product);
  mpz_clear(rq_product);
  mpz_clear(rp_exp);
  mpz_clear(rq_exp);

}

void df_new_Ph_decryption(mpz_t encryption_msg_split_array[][255], 
                          mpz_t plaintext, int previous_counter, 
                          mpz_symmetric_key *mpz_keys_ptr, 
                          FILE *decrypt_time_file_handle)
{

  int i, j, cmp, base=10;
  long seed;
  gmp_randstate_t state;
  int d;
  mpz_t p, q, rp, rq, mpq;
  mpz_t sum, rp_product, rq_product, rp_exp, rq_exp, rp_inverse, rq_inverse;
  mpz_t cipher[255];
  mpz_t sum1, sum2, product;

  /* Variables for chinese remainder theorem*/
  mpz_t m1, m2, M1, M2, M1_inverse, M2_inverse;
  mpz_t plain_text[255];

  mpz_init(mpq);
  mpz_init(p);
  mpz_init(q);
  mpz_init(rp);
  mpz_init(rq);

  for(i=0;i<255;i++){
    mpz_init(plain_text[i]);
    mpz_init(cipher[i]);
  }

  mpz_init(sum);
  mpz_init(sum1);
  mpz_init(sum2);
  mpz_init(rp_product);
  mpz_init(rq_product);
  mpz_init(rp_exp);
  mpz_init(rq_exp);
  mpz_init(rp_inverse);
  mpz_init(rq_inverse);
  mpz_init(product);

  mpz_init(m1);
  mpz_init(m2);
  mpz_init(M1);
  mpz_init(M2);
  mpz_init(M1_inverse);
  mpz_init(M2_inverse);

  gmp_randinit_default(state);
  dec_count = dec_count +1;

  /*Symmetric keys*/
  mpz_set(p, mpz_keys_ptr->p);
  mpz_set(q, mpz_keys_ptr->q);
  mpz_set(rp, mpz_keys_ptr->rp);
  mpz_set(rq, mpz_keys_ptr->rq);
  mpz_set(mpq, mpz_keys_ptr->mpq);
  d = mpz_keys_ptr->d;

  /* Computing inverse of rp and rq*/
  mpz_invert(rp_inverse, rp, p);
  mpz_invert(rq_inverse, rq, q);
  gmp_printf(" the rp inverse:%Zd\n", rp_inverse);
  gmp_printf(" the rq inverse:%Zd\n", rq_inverse);


  /* Setting variables for  Chinese remainder theorem*/
  mpz_set(m1, p);
  mpz_set(m2, q);
  mpz_set(M1, m2);
  mpz_set(M2, m1);
  mpz_invert(M1_inverse, M1, m1);
  mpz_invert(M2_inverse, M2, m2);

  /* Sum up all the incomming messages*/
  mpz_set_str(sum, "0", 10);
  for(i=0;i<(d*2);i++){
    mpz_set_str(sum, "0", 10);
    for(j=0;j<previous_counter;j++){
      mpz_add(sum, sum, encryption_msg_split_array[j][i]);
    }
    mpz_set(cipher[i], sum);
  }

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);


  /*Decryption*/
  for(i=0, j=1;i<(d*2);i=i+2, j++){
    mpz_pow_ui(rp_exp, rp_inverse, j);
    mpz_mul(rp_product, cipher[i], rp_exp);
    mpz_mod(plain_text[i], rp_product, p);
    mpz_add(sum1, sum1, plain_text[i]);
    mpz_mod(sum1, sum1, p);
    mpz_pow_ui(rq_exp, rq_inverse, j);
    mpz_mul(rq_product, cipher[i+1], rq_exp);
    mpz_mod(plain_text[i+1], rq_product, q);
    mpz_add(sum2, sum2, plain_text[i+1]);
    mpz_mod(sum2, sum2, q);
  }

  /* Computing Chinese remainder theorem*/
  mpz_mul(product, M1, M1_inverse);
  mpz_mul(sum1, sum1, product);
  mpz_mul(product, M2, M2_inverse);
  mpz_mul(sum2, sum2, product);
  mpz_add(sum, sum1, sum2);
  mpz_mod(sum, sum, mpq);
  mpz_set(plaintext, sum);

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);

  dec_sum = dec_sum+total_usecs;              
  dec_avg = dec_sum/dec_count;

  printf("TOTAL DECRYPTION TIME : %d uSec", total_usecs);
  //gmp_fprintf(decrypt_time_file_handle, "p: %Zd ", p );
  //gmp_fprintf(decrypt_time_file_handle, "rp: %Zd ", rp );
  gmp_fprintf(decrypt_time_file_handle, "Message: %Zd", plaintext );
  fprintf(decrypt_time_file_handle, " Time: %d uSec sum: %d usec avg: %d usec\n", 
          total_usecs, dec_sum, dec_avg);

  mpz_clear(mpq);
  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(rp);
  mpz_clear(rq);

  for(i=0;i<255;i++){
    mpz_clear(plain_text[i]);
    mpz_clear(cipher[i]);
  }

  mpz_clear(sum);
  mpz_clear(sum1);
  mpz_clear(sum2);
  mpz_clear(rp_product);
  mpz_clear(rq_product);
  mpz_clear(rp_exp);
  mpz_clear(rq_exp);
  mpz_clear(rp_inverse);
  mpz_clear(rq_inverse);
  mpz_clear(product);

  mpz_clear(m1);
  mpz_clear(m2);
  mpz_clear(M1);
  mpz_clear(M2);
  mpz_clear(M1_inverse);
  mpz_clear(M2_inverse);

}


void mmh_encryption(mpz_t msg, char sender_msg_split_buff[], int message_id, 
                    int message_split_id, int message_split_id_sum, 
                    mpz_symmetric_key *mpz_keys_ptr, FILE *encrypt_time_file_handle)
{

  int cmp;
  long seed;
  mpz_t p, q, mpq;
  mpz_t a, random, product, cipher, r, zero;

  size_t *countp=NULL;
  int  buffappend=0;
  int header;
  int zerocmp;
  int encryption_msg_split_size;
  size_t size = sizeof(char);
  size_t numb = 8*size - 0;

  mpz_init_set_str(zero, "0", 10);
  mpz_init(p);
  mpz_init(q);
  mpz_init(mpq);
  mpz_init(a);
  mpz_init(r);
  mpz_init(random);
  mpz_init(product);
  mpz_init(cipher);

  gmp_randstate_t state;
  gmp_randinit_default(state);

  //enc_count = enc_count+1;
  
  /* Getting the encryption keys */
  mpz_set(p, mpz_keys_ptr->p);
  mpz_set(mpq, mpz_keys_ptr->mpq);

  seed = get_seed();
  gmp_randseed_ui(state, seed);

  mpz_urandomm(r,state,p);
  mpz_mul(product, r, p);
  mpz_add(a, msg, product);

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  /* Encryption*/
  mpz_mod(cipher, a, mpq);

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);

  enc_msg_split_id = enc_msg_split_id + message_split_id;
  enc_sum = enc_sum + total_usecs;  
  fprintf(encrypt_time_file_handle, "Message id: %d Message_split_id: %d Time: %d uSec Sum: %d usec",message_id, message_split_id, total_usecs, enc_sum);
  if(enc_msg_split_id == message_split_id_sum){
    enc_msg_split_id=0;
    enc_count++;
    enc_avg_sum = enc_avg_sum + enc_sum;
    enc_avg = enc_avg_sum/enc_count;
    enc_sum =0;
  }
  fprintf(encrypt_time_file_handle, "Avg: %d usec\n",enc_avg);
  

  int *sender_msg_split_int = (int *)sender_msg_split_buff;
  short int *sender_msg_split_short_int = (short int *)sender_msg_split_buff;
  sender_msg_split_buff[0]=message_id;
  sender_msg_split_buff[1]=message_split_id;
  sender_msg_split_short_int[1]=message_split_id_sum;

  header =4;
  buffappend=4;
  encryption_msg_split_size = (mpz_sizeinbase(cipher, 2) + numb-1) / numb;
    
  sender_msg_split_int[1]=encryption_msg_split_size;
  zerocmp = mpz_cmp(cipher, zero);
  if(zerocmp == 0){
    sender_msg_split_buff[header+buffappend] =0;
  }
  else{
    mpz_export(sender_msg_split_buff+header+buffappend, countp, 
               1, 1, 0, 0, cipher);
  }
  mpz_clear(zero);
  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(mpq);
  mpz_clear(a);
  mpz_clear(r);
  mpz_clear(random);
  mpz_clear(product);
  mpz_clear(cipher);
}

void mmh_decryption(mpz_t encryption_msg_split_array[][255], mpz_t plaintext, 
                    int previous_counter, mpz_symmetric_key *mpz_keys_ptr,
                    FILE *decrypt_time_file_handle)
{

  mpz_t p, cipher, sum;

  mpz_init(p);
  mpz_init(cipher);
  mpz_init(sum);
  
  dec_count=dec_count +1;

  /* Symmetric Keys */
  mpz_set(p,mpz_keys_ptr->p);  

  /* Sum up all the incomming messages*/
  mpz_set_str(sum, "0", 10);
  for(j=0;j<previous_counter;j++){
    mpz_add(sum, sum, encryption_msg_split_array[j][0]);
  }
  mpz_set(cipher, sum);

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  /* Decryption */
  mpz_mod(plaintext, cipher, p);

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);
  dec_sum = dec_sum+total_usecs;
  dec_avg = dec_sum/dec_count;

  printf("TOTAL DECRYPTION TIME : %d uSec", total_usecs);
  //gmp_fprintf(decrypt_time_file_handle, "p value: %Zd ", p);
  gmp_fprintf(decrypt_time_file_handle, "Message: %Zd", plaintext );
  fprintf(decrypt_time_file_handle, " Time:%d uSec sum: %d usec avg: %d usec\n", 
          total_usecs, dec_sum, dec_avg);

  mpz_clear(p);
  mpz_clear(cipher);
  mpz_clear(sum);

}


void df_field_encryption(mpz_t msg, char sender_msg_split_buff[], 
                         int message_id, int message_split_id, 
                         int message_split_id_sum, 
                         mpz_symmetric_key *mpz_keys_ptr, 
                         FILE *encrypt_time_file_handle)
{

  int cmp;
  long seed;
  mpz_t a, b, b_inverse, product, cipher, p, q, mpq, r, zero;

  size_t *countp=NULL;
  int  buffappend=0;
  int header;
  int zerocmp;
  int encryption_msg_split_size;
  size_t size = sizeof(char);
  size_t numb = 8*size - 0;

  mpz_init_set_str(zero, "0", 10);
  mpz_init(a);
  mpz_init(b);
  mpz_init(p);
  mpz_init(q);
  mpz_init(mpq);
  mpz_init(r);
  mpz_init(b_inverse);
  mpz_init(product);
  mpz_init(cipher);

  gmp_randstate_t state;
  gmp_randinit_default(state);

  /* Getting the symmetric keys */
  mpz_set(p, mpz_keys_ptr->p);
  mpz_set(mpq, mpz_keys_ptr->mpq);


  seed = get_seed();
  gmp_randseed_ui(state, seed);

  //enc_count = enc_count+1;

  mpz_urandomm(r,state,p);
  mpz_mul(a, msg, r);
  mpz_invert(b_inverse, r, p);
  mpz_mod(a, a, p); 
  
  printf("****Encrypting with field encryption***\n");

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  /* Encryption*/
  mpz_mul(product, a, b_inverse);
  mpz_mod(cipher, product, mpq);

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);

  enc_msg_split_id = enc_msg_split_id + message_split_id;
  enc_sum = enc_sum + total_usecs;  
  fprintf(encrypt_time_file_handle, "Message id: %d  message_split_id: %d Time: %d uSec Sum: %d usec",message_id, message_split_id, total_usecs, enc_sum);
  if(enc_msg_split_id == message_split_id_sum){
    enc_msg_split_id=0;
    enc_count++;
    enc_avg_sum = enc_avg_sum + enc_sum;
    enc_avg = enc_avg_sum/enc_count;
    enc_sum =0;
  }
  fprintf(encrypt_time_file_handle, "Avg: %d usec\n",enc_avg);
  printf("****Encrypted with field encryption***\n");

  

  int *sender_msg_split_int = (int *)sender_msg_split_buff;
  short int *sender_msg_split_short_int = (short int *)sender_msg_split_buff;
  sender_msg_split_buff[0]=message_id;
  sender_msg_split_buff[1]=message_split_id;
  sender_msg_split_short_int[1]=message_split_id_sum;

  header =4;
  buffappend=4;
  encryption_msg_split_size = (mpz_sizeinbase(
                               cipher, 2) +
                               numb-1) / numb;
  sender_msg_split_int[1]=encryption_msg_split_size;
  zerocmp = mpz_cmp(cipher, zero);
  if(zerocmp == 0){
    sender_msg_split_buff[header+buffappend] =0;
  }
  else{
    mpz_export(sender_msg_split_buff+header+buffappend, countp, 
                 1, 1, 0, 0, cipher);
  }

  mpz_clear(zero);
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(mpq);
  mpz_clear(r);
  mpz_clear(b_inverse);
  mpz_clear(product);
  mpz_clear(cipher);

}

void df_field_decryption(mpz_t encryption_msg_split_array[][255], 
                         mpz_t plaintext, 
                         int previous_counter, mpz_symmetric_key *mpz_keys_ptr,
                         FILE *decrypt_time_file_handle)
{

  int cmp;
  long seed;
  mpz_t mpq, p, a, b, b_inverse, product, sum, cipher;
  gmp_randstate_t state;

  mpz_init(p);
  mpz_init(mpq);
  mpz_init(a);
  mpz_init(b);
  mpz_init(b_inverse);
  mpz_init(product);
  mpz_init(sum);
  mpz_init(cipher);

  gmp_randinit_default(state);

  dec_count = dec_count+1;

  /* Symmetric Keys*/
  mpz_set(p,mpz_keys_ptr->p);  
  mpz_set(mpq,mpz_keys_ptr->mpq);  

  /* Sum up all the incomming messages*/
  mpz_set_str(sum, "0", 10);
  for(j=0;j<previous_counter;j++){
    mpz_add(sum, sum, encryption_msg_split_array[j][0]);
  }
  mpz_set(cipher, sum);



  seed = get_seed();
  gmp_randseed_ui(state, seed);

  mpz_t btemp;
  mpz_init(btemp);

  mpz_urandomm(b, state, mpq);
  mpz_set(btemp, b);
  mpz_mul(a, cipher, btemp);
  mpz_mod(a, a, mpq);
  mpz_invert(b_inverse, b, mpq);

  /* Get gettimeofday() to get start time */
  gettimeofday(&start_time, NULL);

  printf("*** The decryption with field***\n");

  /* Decryption*/
  mpz_mul(product, a, b_inverse);
  mpz_mod(plaintext, product, p);

  /* Now call gettimeofday() to get end time */
  gettimeofday(&end_time, NULL);  /* after time */

  /* Get the execution time */
  total_usecs = (end_time.tv_sec-start_time.tv_sec) * 1000000 +
                (end_time.tv_usec-start_time.tv_usec);
  dec_sum = dec_sum+total_usecs;
  dec_avg = dec_sum/dec_count;

  printf("TOTAL DECRYPTION TIME : %d uSec", total_usecs);
  gmp_fprintf(decrypt_time_file_handle, "Message: %Zd ", plaintext );
  fprintf(decrypt_time_file_handle, "Time:%d uSec sum: %d usec avg: %d usec\n", 
          total_usecs, dec_sum, dec_avg);

  mpz_clear(p);
  mpz_clear(mpq);
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(b_inverse);
  mpz_clear(product);
  mpz_clear(sum);
  mpz_clear(cipher);

}


void df_add_mul_buff_to_mpz_array(char receivebuff[], 
                                  mpz_t encryption_msg_split_array[],
                                  int encryption_split_number)
{

  int *int_buff = (int *)receivebuff;
  int header= (encryption_split_number * 4)+ 4;
  int buffappend = 0;
  int i;
  int encryption_msg_split_size;

  for(i=0;i<encryption_split_number;i++){
    encryption_msg_split_size = int_buff[i+1];
    mpz_import(encryption_msg_split_array[i], 
               encryption_msg_split_size, 1, 1, 0, 0, 
               receivebuff+header+buffappend);

    buffappend = buffappend +encryption_msg_split_size;

    gmp_printf("The imported message at the receiver:%Zd\n", 
               encryption_msg_split_array[i]);
  }

}

void df_new_buff_to_mpz_array(char receivebuff[], 
                              mpz_t encryption_msg_split_array[], 
                              int encryption_split_number)
{

  int *int_buff = (int *)receivebuff;
  int header= ((encryption_split_number*2) * 4)+ 4;
  int buffappend = 0;
  int i;
  int encryption_msg_split_size;

  for(i=0;i<(encryption_split_number*2);i++){
    encryption_msg_split_size = int_buff[i+1];
    mpz_import(encryption_msg_split_array[i], 
               encryption_msg_split_size, 1, 1, 0, 0, 
               receivebuff+header+buffappend);

    buffappend = buffappend +encryption_msg_split_size;

    gmp_printf("The imported message at the receiver:%Zd\n", 
               encryption_msg_split_array[i]);
  }
}

void df_field_mmh_buff_to_mpz_array(char receivebuff[], 
                                    mpz_t encryption_msg_split_array[])
{
  int *int_buff = (int *)receivebuff;
  int header= 4;
  int buffappend = 4;
  int encryption_msg_split_size;

  encryption_msg_split_size = int_buff[1];
  mpz_import(encryption_msg_split_array[0], 
             encryption_msg_split_size, 1, 1, 0, 0, 
             receivebuff+header+buffappend);


  gmp_printf("The imported message at the receiver:%Zd\n", 
               encryption_msg_split_array[0]);
}


void encrypt(mpz_t splitmsg, char sender_msg_split_buff[], int message_id, 
             int message_split_id, int message_split_id_sum, 
             mpz_symmetric_key *mpz_keys_ptr)
{
  FILE *encrypt_time_file_handle;
  encrypt_time_file_handle = fopen(manets_log, "a+");

  if(mpz_keys_ptr->Ph == DA){
    printf(" Encrypting with DA field\n");
    df_add_mul_Ph_encryption(splitmsg, sender_msg_split_buff, message_id, 
                             message_split_id, message_split_id_sum, 
                             mpz_keys_ptr, 
                             encrypt_time_file_handle);
    printf(" Encrypted with DA field\n");
  }

  if(mpz_keys_ptr->Ph == DN){
    printf(" Encrypting with DN field\n");
    df_new_Ph_encryption(splitmsg, sender_msg_split_buff, message_id, 
                         message_split_id, message_split_id_sum, mpz_keys_ptr, 
                         encrypt_time_file_handle);
    printf(" Encrypted with DN field\n");
  }

  if(mpz_keys_ptr->Ph == MMH){
    printf(" Encrypting with mmh field\n");
    mmh_encryption(splitmsg, sender_msg_split_buff, message_id, 
                   message_split_id, message_split_id_sum, mpz_keys_ptr, 
                   encrypt_time_file_handle);
    printf(" Encrypted with mmh field\n");
  }

  if(mpz_keys_ptr->Ph == DF){
    printf(" Encrypting with DF field\n");
    df_field_encryption(splitmsg, sender_msg_split_buff, message_id, 
                        message_split_id, message_split_id_sum, mpz_keys_ptr, 
                        encrypt_time_file_handle);
    printf(" Encrypted with DF field\n");
  }
  
  fclose(encrypt_time_file_handle);

}


void decrypt(mpz_t encryption_msg_split_array[][255], mpz_t plaintext, 
             int previous_counter, mpz_symmetric_key *mpz_keys_ptr)
{

   FILE *decrypt_time_file_handle;
   decrypt_time_file_handle = fopen(manets_log, "a+");


  if(mpz_keys_ptr->Ph == DA){
    printf(" The Encryption is PH addive Multiplicative\n");
    df_add_mul_Ph_decryption(encryption_msg_split_array, plaintext, 
                             previous_counter,
                             mpz_keys_ptr, decrypt_time_file_handle);
  }  
  if(mpz_keys_ptr->Ph == DN){
    printf(" The Encryption is New PH\n");
    df_new_Ph_decryption(encryption_msg_split_array, plaintext, 
                         previous_counter, 
                         mpz_keys_ptr, decrypt_time_file_handle);
  }  
  if(mpz_keys_ptr->Ph == MMH){
    printf(" The Encryption is MMH\n");
    mmh_decryption(encryption_msg_split_array, plaintext, previous_counter, 
                   mpz_keys_ptr, decrypt_time_file_handle);
  }  
  if(mpz_keys_ptr->Ph == DF){
    printf(" Entering the field decryption\n");
    printf(" The Encryption is PH Mixed Multiplicative\n");
    df_field_decryption(encryption_msg_split_array, plaintext, previous_counter, 
                        mpz_keys_ptr, decrypt_time_file_handle);
    printf(" Done with field decryption\n");
  }  

  fclose(decrypt_time_file_handle);

}


void buff_to_mpz_array(char receivebuff[], mpz_t encryption_msg_split_array[], 
                                  int encryption_split_number, 
                                  enc_scheme encryption_scheme)
{

  if(encryption_scheme == DA){
    df_add_mul_buff_to_mpz_array(receivebuff, encryption_msg_split_array, 
                                 encryption_split_number);
  }
  if(encryption_scheme == DN){
    df_new_buff_to_mpz_array(receivebuff, encryption_msg_split_array, 
                             encryption_split_number);
  }
  if(encryption_scheme == MMH || encryption_scheme == DF){
    df_field_mmh_buff_to_mpz_array(receivebuff, encryption_msg_split_array);
  }
}

void set_encryption_scheme(enc_scheme encryption_scheme){

  /* Reading the shared Memory*/
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    sk_ptr ->Ph = encryption_scheme;
  sem_post(enc_mutex);
  sem_close(enc_mutex);
  munmap(sk_ptr, sizeof(symmetric_key));

}

enc_scheme get_encryption_scheme(){

  /* Reading the shared Memory*/
  enc_scheme encryption_scheme;
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
   encryption_scheme = sk_ptr->Ph;
  sem_post(enc_mutex);
  sem_close(enc_mutex);
  munmap(sk_ptr, sizeof(symmetric_key));
  return encryption_scheme;

}


void set_df_add_mul_ph_keys(){
  int cmp, comp;
  int d;
  mpz_t md, m, r, r_inverse;
  long seed;
  gmp_randstate_t state;

  mpz_init(md);
  mpz_init(m);
  mpz_init(r);
  mpz_init(r_inverse);


  printf("manets_simulation> Enter the r value : ");
  mpz_inp_str(r, NULL, 10);

  printf("manets_simulation> Enter m' value : ");
  mpz_inp_str(md, NULL, 10);

  printf("manets_simulation> Enter the m value : ");
  mpz_inp_str(m, NULL, 10);

  printf("manets_simulation> Enter d value : ");
  scanf("%d", &d);
  
  gmp_randinit_default(state);

  seed = get_seed();
  gmp_randseed_ui(state, seed);

  /* Opening the shared Memory and setting the key values in Shared Memory */
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    long size = sizeof(char);
    long numb = 8*size - 0;

    sk_ptr->r_count = (mpz_sizeinbase (r, 2) + numb-1) / numb;
    sk_ptr->md_count = (mpz_sizeinbase (md, 2) + numb-1) / numb;
    sk_ptr->m_count = (mpz_sizeinbase (m, 2) + numb-1) / numb;

    mpz_export(sk_ptr->r_buff, NULL, 1, 1, 0, 0, r);
    mpz_export(sk_ptr->md_buff, NULL, 1, 1, 0, 0, md);
    mpz_export(sk_ptr->m_buff, NULL, 1, 1, 0, 0, m);
    sk_ptr->d =d;
  sem_post(enc_mutex);
  sem_close(enc_mutex);

  mpz_clear(md);
  mpz_clear(m);
  mpz_clear(r);
  mpz_clear(r_inverse);
  munmap(sk_ptr, sizeof(symmetric_key));

}

void set_df_new_ph_keys(){
  int d;
  mpz_t p, q, m, rp, rq, p_prime, q_prime;

  mpz_init(p);
  mpz_init(rp);
  mpz_init(p_prime);
  mpz_init(q);
  mpz_init(rq);
  mpz_init(q_prime);
  mpz_init(m);

  printf("manets_simulation> Enter p value : ");
  mpz_inp_str(p_prime, NULL, 10);

  printf("manets_simulation> Enter rp value : ");
  mpz_inp_str(rp, NULL, 10);

  printf("manets_simulation> Enter q value : ");
  mpz_inp_str(q_prime, NULL, 10);

  printf("manets_simulation> Enter rq value : ");
  mpz_inp_str(rq, NULL, 10);

  printf("manets_simulation> Enter d value : ");
  scanf("%d", &d);

  mpz_mul(m, p_prime, q_prime);

  /* Opening the shared Memory and setting the key values in Shared Memory */
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    long size = sizeof(char);
    long numb = 8*size - 0;

    sk_ptr->p_count = (mpz_sizeinbase (p_prime, 2) + numb-1) / numb;
    sk_ptr->q_count = (mpz_sizeinbase (q_prime, 2) + numb-1) / numb;
    sk_ptr->rp_count = (mpz_sizeinbase (rp, 2) + numb-1) / numb;
    sk_ptr->rq_count = (mpz_sizeinbase (rq, 2) + numb-1) / numb;
    sk_ptr->mpq_count = (mpz_sizeinbase (m, 2) + numb-1) / numb;

    mpz_export(sk_ptr->p_buff, NULL, 1, 1, 0, 0, p_prime);
    mpz_export(sk_ptr->q_buff, NULL, 1, 1, 0, 0, q_prime);
    mpz_export(sk_ptr->rp_buff, NULL, 1, 1, 0, 0, rp);
    mpz_export(sk_ptr->rq_buff, NULL, 1, 1, 0, 0, rq);
    mpz_export(sk_ptr->mpq_buff, NULL, 1, 1, 0, 0, m);
    sk_ptr->d=d;
  sem_post(enc_mutex);
  sem_close(enc_mutex);

  mpz_clear(p);
  mpz_clear(rp);
  mpz_clear(p_prime);
  mpz_clear(q);
  mpz_clear(rq);
  mpz_clear(q_prime);
  mpz_clear(m);
  munmap(sk_ptr, sizeof(symmetric_key));
}

void set_df_field_mmh_ph_keys(){
  mpz_t p, q, m, p_prime, q_prime, r;

  mpz_init(p);
  mpz_init(p_prime);
  mpz_init(q);
  mpz_init(q_prime);
  mpz_init(m);
  mpz_init(r);

  printf("manets_simulation> Enter p value : ");
  mpz_inp_str(p_prime, NULL, 10);

  printf("manets_simulation> Enter q value : ");
  mpz_inp_str(q_prime, NULL, 10);

  mpz_mul(m, p_prime, q_prime);


  /* Opening the shared Memory and setting the key values in Shared Memory */
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);

    long size = sizeof(char);
    long numb = 8*size - 0;

    sk_ptr->p_count = (mpz_sizeinbase (p_prime, 2) + numb-1) / numb;
    sk_ptr->q_count = (mpz_sizeinbase (q_prime, 2) + numb-1) / numb;
    sk_ptr->mpq_count = (mpz_sizeinbase (m, 2) + numb-1) / numb;

    mpz_export(sk_ptr->p_buff, NULL, 1, 1, 0, 0, p_prime);
    mpz_export(sk_ptr->q_buff, NULL, 1, 1, 0, 0, q_prime);
    mpz_export(sk_ptr->mpq_buff, NULL, 1, 1, 0, 0, m);

  sem_post(enc_mutex);
  sem_close(enc_mutex);

  mpz_clear(p);
  mpz_clear(p_prime);
  mpz_clear(q);
  mpz_clear(q_prime);
  mpz_clear(m);
  mpz_clear(r);
  munmap(sk_ptr, sizeof(symmetric_key));
}

void get_various_encryption_keys(mpz_symmetric_key *mpz_keys_ptr){

  symmetric_key *sk_ptr;

  mpz_init(mpz_keys_ptr->r);
  mpz_init(mpz_keys_ptr->md);
  mpz_init(mpz_keys_ptr->m);
  mpz_init(mpz_keys_ptr->p);
  mpz_init(mpz_keys_ptr->q);
  mpz_init(mpz_keys_ptr->rp);
  mpz_init(mpz_keys_ptr->rq);
  mpz_init(mpz_keys_ptr->mpq);

  /* Reading from Shared memory */
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);

    printf(" importing The r key \n");
    mpz_import(mpz_keys_ptr->r, sk_ptr->r_count, 1, 1, 0, 0, sk_ptr->r_buff);
    gmp_printf(" The r key :%Zd", mpz_keys_ptr->r);
    mpz_import(mpz_keys_ptr->md, sk_ptr->md_count, 1, 1, 0, 0, sk_ptr->md_buff);
    mpz_import(mpz_keys_ptr->m, sk_ptr->m_count, 1, 1, 0, 0, sk_ptr->m_buff);

    mpz_import(mpz_keys_ptr->p, sk_ptr->p_count, 1, 1, 0, 0, sk_ptr->p_buff);
    mpz_import(mpz_keys_ptr->rp, sk_ptr->rp_count, 1, 1, 0, 0, sk_ptr->rp_buff);
    mpz_import(mpz_keys_ptr->q, sk_ptr->q_count, 1, 1, 0, 0, sk_ptr->q_buff);
    mpz_import(mpz_keys_ptr->rq, sk_ptr->rq_count, 1, 1, 0, 0, sk_ptr->rq_buff);
    mpz_import(mpz_keys_ptr->mpq, sk_ptr->mpq_count, 1, 1, 0, 0, 
               sk_ptr->mpq_buff);
    mpz_keys_ptr->d = sk_ptr->d;
    mpz_keys_ptr->Ph = sk_ptr->Ph;

  sem_post(enc_mutex);
  sem_close(enc_mutex);
  munmap(sk_ptr, sizeof(symmetric_key));

}

void set_df_mul_ph_rd_keys(mpz_t value){
  int cmp=0;
  int d;
  mpz_t r,md,m;
  mpz_init(r);
  mpz_init(md);
  mpz_init(m);

  printf("manets_simulation> ");
  if( mpz_cmp_ui(value, 11) == 0){
    printf(" Enter the r value : ");
    mpz_inp_str(r, NULL, 10);
  }

  if( mpz_cmp_ui(value, 12) == 0){
    printf(" Enter the m' value : ");
    mpz_inp_str(md, NULL, 10);
  }

  if( mpz_cmp_ui(value, 13) == 0){
    printf(" Enter the m value : ");
    mpz_inp_str(m, NULL, 10);
  }
  
  if( mpz_cmp_ui(value, 14) == 0){
      printf(" Enter the d value : ");
      scanf("%d", &d);
  }



  /* Opening the shared Memory and setting the r or d key in Shared Memory */
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    long size = sizeof(char);
    long numb = 8*size - 0;

    if(mpz_cmp_ui(value, 11) == 0){ 
      sk_ptr->r_count = (mpz_sizeinbase (r, 2) + numb-1) / numb;
      mpz_export(sk_ptr->r_buff, NULL, 1, 1, 0, 0, r);
    }  
    if(mpz_cmp_ui(value, 12) == 0){ 
      sk_ptr->md_count = (mpz_sizeinbase (md, 2) + numb-1) / numb;
      mpz_export(sk_ptr->md_buff, NULL, 1, 1, 0, 0, md);
    }  

    if(mpz_cmp_ui(value, 13) == 0){ 
      sk_ptr->m_count = (mpz_sizeinbase (m, 2) + numb-1) / numb;
      mpz_export(sk_ptr->m_buff, NULL, 1, 1, 0, 0, m);
    }  

    if(mpz_cmp_ui(value, 14) == 0){ 
      sk_ptr->d = d;
    }

  sem_post(enc_mutex);
  sem_close(enc_mutex);

  mpz_clear(r);
  munmap(sk_ptr, sizeof(symmetric_key));

}

void set_df_new_ph_rp_p_d_keys(mpz_t value){
  int cmp=0;
  int d;
  mpz_t rp, rq, p, mpq, q;

  mpz_init(rp);
  mpz_init(rq);
  mpz_init(p);
  mpz_init(q);
  mpz_init(mpq);

  printf("manets_simulation> ");

  if(mpz_cmp_ui(value, 11)==0){ 
    printf(" Enter the p value : ");
    mpz_inp_str(p, NULL, 10);
    gmp_printf(" The p value: %Zd\n", p);
  }

  if(mpz_cmp_ui(value, 12)==0){ 
    printf(" Enter the rp value : ");
    mpz_inp_str(rp, NULL, 10);
    gmp_printf(" The rp value: %Zd\n", rp);
  }

  if(mpz_cmp_ui(value, 13)==0){ 
    printf(" Enter the q value : ");
    mpz_inp_str(q, NULL, 10);
    gmp_printf(" The q value: %Zd\n", q);
  }

  if(mpz_cmp_ui(value, 14)==0){ 
    printf(" Enter the rq value : ");
    mpz_inp_str(rq, NULL, 10);
    gmp_printf(" The rq value: %Zd\n", rq);
  }

  if(mpz_cmp_ui(value, 15)==0){ 
    printf(" Enter the d value : ");
    scanf("%d", &d);
  }


  /* Opening the shared Memory and setting the r or d key in Shared Memory */
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    long size = sizeof(char);
    long numb = 8*size - 0;

    if(mpz_cmp_ui(value, 11)==0){ 
      sk_ptr->p_count = (mpz_sizeinbase (p, 2) + numb-1) / numb;
      mpz_export(sk_ptr->p_buff, NULL, 1, 1, 0, 0, p);
      mpz_import(q, sk_ptr->q_count, 1, 1, 0, 0, sk_ptr->q_buff);
      mpz_mul(mpq, p, q);
      sk_ptr->mpq_count = (mpz_sizeinbase (mpq, 2) + numb-1) / numb;
      mpz_export(sk_ptr->mpq_buff, NULL, 1, 1, 0, 0, mpq);
    }  

    if(mpz_cmp_ui(value, 12)==0){ 
      sk_ptr->rp_count = (mpz_sizeinbase (rp, 2) + numb-1) / numb;
      mpz_export(sk_ptr->rp_buff, NULL, 1, 1, 0, 0, rp);
    }  

    if(mpz_cmp_ui(value, 13)==0){ 
      sk_ptr->q_count = (mpz_sizeinbase (q, 2) + numb-1) / numb;
      mpz_export(sk_ptr->q_buff, NULL, 1, 1, 0, 0, q);
      mpz_import(p, sk_ptr->p_count, 1, 1, 0, 0, sk_ptr->p_buff);
      mpz_mul(mpq, p, q);
      sk_ptr->mpq_count = (mpz_sizeinbase (mpq, 2) + numb-1) / numb;
      mpz_export(sk_ptr->mpq_buff, NULL, 1, 1, 0, 0, mpq);
    }  

    if(mpz_cmp_ui(value, 14)==0){ 
      sk_ptr->rq_count = (mpz_sizeinbase (rq, 2) + numb-1) / numb;
      mpz_export(sk_ptr->rq_buff, NULL, 1, 1, 0, 0, rq);
    }  

    if(mpz_cmp_ui(value, 15)==0){ 
      sk_ptr->d = d;
    }

  sem_post(enc_mutex);
  sem_close(enc_mutex);

  mpz_clear(rp);
  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(mpq);
  munmap(sk_ptr, sizeof(symmetric_key));

}

void set_df_field_mmh_ph_p_r_keys(mpz_t value){
  int cmp=0;
  mpz_t p, mpq, q, r;

  mpz_init(p);
  mpz_init(q);
  mpz_init(r);
  mpz_init(mpq);

  printf("manets_simulation> ");

  if(mpz_cmp_ui(value, 11)==0){ 
    printf(" Enter the p value : ");
    mpz_inp_str(p, NULL, 10);
    gmp_printf(" The p value: %Zd\n", p);
  }


  /* Opening the shared Memory and setting the r or d key in Shared Memory */
  symmetric_key *sk_ptr;
  sk_ptr = get_symmetric_key();

  enc_mutex = sem_open(ENC_KEYS_TABLE_LOCK, 0);
  sem_wait(enc_mutex);
    long size = sizeof(char);
    long numb = 8*size - 0;

    if(mpz_cmp_ui(value, 11)==0){ 
      sk_ptr->p_count = (mpz_sizeinbase (p, 2) + numb-1) / numb;
      mpz_export(sk_ptr->p_buff, NULL, 1, 1, 0, 0, p);
      mpz_import(q, sk_ptr->q_count, 1, 1, 0, 0, sk_ptr->q_buff);
      mpz_mul(mpq, p, q);
      sk_ptr->mpq_count = (mpz_sizeinbase (mpq, 2) + numb-1) / numb;
      mpz_export(sk_ptr->mpq_buff, NULL, 1, 1, 0, 0, mpq);
    }  

  sem_post(enc_mutex);
  sem_close(enc_mutex);

  mpz_clear(p);
  mpz_clear(q);
  mpz_clear(r);
  mpz_clear(mpq);
  munmap(sk_ptr, sizeof(symmetric_key));

}

void init_route_table(){
  /* Open the shared memory and initialize it*/

  route_table *rt_ptr;
  rt_ptr = get_route_table();

  node_mutex = sem_open(ROUTER_TABLE_LOCK, 0);

  /* Initialize the route table */
  sem_wait(node_mutex);
    rt_ptr->node_counter = 0;
    rt_ptr->group_counter = 1;
    rt_ptr->level_counter = 1;
    rt_ptr->path_counter = 1;
  sem_post(node_mutex);
  sem_close(node_mutex);
  munmap(rt_ptr, sizeof(route_table));

}

void add_node(int pid){

  /* Open the shared memory*/

  route_table *rt_ptr;
  rt_ptr = get_route_table();

  node_mutex = sem_open(ROUTER_TABLE_LOCK, 0);
  sem_wait(node_mutex);

  switch(rt_ptr->node_counter){

    case STARTING_NODE:

      /*The routing table set for the sender*/
      rt_ptr->pids[rt_ptr->node_counter]=pid;
      rt_ptr->levels[rt_ptr->node_counter]=0;
      rt_ptr->groups[rt_ptr->node_counter]=0;
      rt_ptr->paths[rt_ptr->node_counter]=0;
      rt_ptr->status[rt_ptr->node_counter]= 1;
      sprintf(rt_ptr->message_queues[rt_ptr->node_counter], "/route.%d", pid);
      rt_ptr->node_counter = rt_ptr->node_counter+1;
      break;

    case (MAX_NODES -1 + STARTING_NODE):  

      /*The routing table set for the receiver*/
      rt_ptr->pids[rt_ptr->node_counter]=pid;
      rt_ptr->levels[rt_ptr->node_counter]=LEVELS+STARTING_LEVEL+1;
      rt_ptr->groups[rt_ptr->node_counter]=STARTING_ID -1;
      rt_ptr->paths[rt_ptr->node_counter]=STARTING_ID -1;
      rt_ptr->status[rt_ptr->node_counter]= 1;
      sprintf(rt_ptr->message_queues[rt_ptr->node_counter], "/route.%d", pid);
      rt_ptr->node_counter = rt_ptr->node_counter+1;
      break;

    default:  

      /*The routing table set for intermediate nodes*/ 
      rt_ptr->pids[rt_ptr->node_counter]=pid;
      rt_ptr->status[rt_ptr->node_counter]= 1;
      sprintf(rt_ptr->message_queues[rt_ptr->node_counter], "/route.%d", pid);

      if(rt_ptr->node_counter
         <= 
         (NODES_PER_GROUP*NEIGHBORING_NODES*rt_ptr->level_counter)
        ){
        rt_ptr->levels[rt_ptr->node_counter]= rt_ptr->level_counter;
      }

      if(rt_ptr->node_counter<= (NODES_PER_GROUP * rt_ptr->group_counter)){
        rt_ptr->groups[rt_ptr->node_counter]=rt_ptr->group_counter;
        rt_ptr->paths[rt_ptr->node_counter]=rt_ptr->path_counter;
      }

      rt_ptr->node_counter = rt_ptr->node_counter+1;
      break;
   
  }

  if(rt_ptr->node_counter > (NODES_PER_GROUP * rt_ptr->group_counter)){
    rt_ptr->group_counter++;
    rt_ptr->path_counter++;

  }

  if(rt_ptr->node_counter > 
     NODES_PER_GROUP*NEIGHBORING_NODES*rt_ptr->level_counter){
    rt_ptr->level_counter = rt_ptr->level_counter+1; 
    rt_ptr->path_counter=1;
  }

  sem_post(node_mutex);
  sem_close(node_mutex);
  munmap(rt_ptr, sizeof(route_table));

}

void get_next_node(int pid, char msg_queue[][100], int neighbouring_groups[], 
                   int *count){

  /*Open the shared memory*/
  route_table *rt_ptr;
  int i, j;
  int level=0, path=0;

  rt_ptr = get_route_table();
  node_mutex = sem_open(ROUTER_TABLE_LOCK, O_CREAT, 0);

  sem_wait(node_mutex);
    for(i=0;rt_ptr->pids[i]!=pid;i++);
    if(rt_ptr->pids[i]==pid){
      level = rt_ptr->levels[i];
      path = rt_ptr->paths[i];
    }
    if(level!=0 && level!= LEVELS){
      j=0;
      *count=0;
      for(i=0;i<MAX_NODES;i++){
        if(rt_ptr->levels[i]==level+1 && rt_ptr->paths[i]==path  
           && rt_ptr->status[i]==1){
          strcpy(msg_queue[j], rt_ptr->message_queues[i]);
          neighbouring_groups[j] = rt_ptr->groups[i];
          j++;
          *count=j;
        }
      }
    }
    else if(level == 0){
      j=0;
      *count=0;
      for(i=0;i<MAX_NODES;i++){
        if(rt_ptr->levels[i]==level+1 && rt_ptr->status[i]==1){
          printf("Found the level");
          strcpy(msg_queue[j], rt_ptr->message_queues[i]);
          neighbouring_groups[j] = rt_ptr->groups[i];
          j++;
          *count=j;
          }
      }
    }  
    else if(level == LEVELS){
      j=0;
      *count=0;
      for(i=0;i<MAX_NODES;i++){
        if(rt_ptr->levels[i]==level+1 && rt_ptr->status[i]==1){
          printf("Found the level");
          strcpy(msg_queue[j], rt_ptr->message_queues[i]);
          neighbouring_groups[j] = rt_ptr->groups[i];
          j++;
          *count=j;
        }
      }    

    }


  sem_post(node_mutex);
  sem_close(node_mutex);
  munmap(rt_ptr, sizeof(route_table));

}

void get_sender_mqueue(char msg_queue[][100]){

  route_table *rt_ptr;
  rt_ptr = get_route_table();

  node_mutex = sem_open(ROUTER_TABLE_LOCK, O_CREAT, 0);
  sem_wait(node_mutex);

    strcpy(msg_queue[0], rt_ptr->message_queues[0]);

  sem_post(node_mutex);
  sem_close(node_mutex);
  munmap(rt_ptr, sizeof(route_table));

}


long get_seed(){
  long seed;

  /* Use /dev/urandom to set the value of seed*/
  FILE *devrandom = fopen("/dev/urandom", "r");
  fread(&seed, sizeof(seed), 1 , devrandom);
  fclose(devrandom);
  return seed;
}

symmetric_key *get_symmetric_key(){
  /* Opening the shared Memory  */

  int sk_d;
  symmetric_key *sk_ptr;

  sk_d = shm_open(ENC_KEYS_TABLE, O_RDWR, 0666);
  sk_ptr = mmap(NULL, sizeof(symmetric_key), PROT_READ|PROT_WRITE, 
                  MAP_SHARED, sk_d, 0);
  close(sk_d);
  return sk_ptr;
}

route_table *get_route_table(){
  /* Opening the shared Memory  */

  int rt_d;
  route_table *rt_ptr;

  rt_d = shm_open(ROUTER_TABLE, O_RDWR, 0666);
  rt_ptr = mmap(NULL, sizeof(route_table), PROT_READ|PROT_WRITE, 
                MAP_SHARED , rt_d, 0);
  close(rt_d);
  return rt_ptr;
}

void set_manets_log(char *log_path){
 manets_log = getenv("MANETS_HOME");
 if(manets_log == NULL)
   strcpy(manets_log,"/home/vaidehi");
 strcat(manets_log,"/");
 strcat(manets_log,log_path);
}
