/* 
  Program :manets_simulation.h
  Author  :Vaidehi 
  Notes   :
    1. Sender knowns the number of neighboring nodes available to them.
    2. Routing protocol ditermines the disjoint paths between the 
       sender and the receiver.
    3. If one of the nodes in the disjoint path dies then the 
       disjoint path is not available.
    4. The number of splits are always lesser than the number of 
       available neighbors to the sender.
    5. Minimum Requirements:SENDER_MSG_SPLIT_COUNT = 1
                            NEIGHBORING_NODES = 2  
                            LEVELS = 1
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <mqueue.h>
#include <gmp.h>
#include <semaphore.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#ifndef SENDER_MSG_SPLIT_COUNT  
# define  SENDER_MSG_SPLIT_COUNT 2 
#endif

# define MAXLEN 1200
# define MAX_BUFF_SIZE 8000 

#ifndef LEVELS
# define LEVELS 2 
#endif

# define NODES_PER_GROUP 1  

#ifndef NEIGHBORING_NODES 
# define NEIGHBORING_NODES 2 
#endif

#ifndef EXP_DIST_MEAN 
# define EXP_DIST_MEAN 50.0 
#endif

# define MAX_NODES (NEIGHBORING_NODES* NODES_PER_GROUP * LEVELS)+2
# define STARTING_LEVEL 0   /* Including Sender */
# define STARTING_PATH 1 
# define STARTING_NODE 0
# define STARTING_ID 1
# define MAX_BUFF_SIZE 8000 

/* Log directory */
# define ENC_LOG_FILE "thesis/log/encryption_exec_time.log"
# define DEC_LOG_FILE "thesis/log/decryption_exec_time.log"

#define ROUTER_TABLE "/route_table_memory"
#define ROUTER_TABLE_LOCK "/router_table_lock" 

#define ENC_KEYS_TABLE "/enc_keys_table_memory"
#define ENC_KEYS_TABLE_LOCK "/enc_keys_table_lock" 

/* Encryption schemes */
typedef enum enc_scheme{
    DA
  , DF
  , DN
  , MMH
} enc_scheme;

/* Structure for the symmetric key shared memory used across the processes*/
typedef struct symmetric_key{
  long r_count;
  long md_count;
  long m_count;
  char r_buff[800];
  char md_buff[800];
  char m_buff[800];
  long d;
  long p_count;
  long q_count;
  long rp_count;
  long rq_count;
  long mpq_count;
  char p_buff[800];
  char q_buff[800];
  char rp_buff[800];
  char rq_buff[800];
  char mpq_buff[800];
  enc_scheme Ph;
}symmetric_key;

/* Process global symmetric key */
typedef struct mpz_symmetric_key{
  mpz_t r;
  mpz_t md;
  mpz_t m;
  long d;
  mpz_t p;
  mpz_t q;
  mpz_t rp;
  mpz_t rq;
  mpz_t mpq;
  enc_scheme Ph;
} mpz_symmetric_key;

/* Structure for the route_table shared memory used across the processes*/
typedef struct route_table{
  int pids[MAX_NODES];
  int levels[MAX_NODES];
  int groups[MAX_NODES];
  unsigned char message_queues[MAX_NODES][100];
  int paths[MAX_NODES];
  int status[MAX_NODES];
  int node_counter;
  int group_counter;
  int level_counter;
  int path_counter;
}route_table;

/* Various mutexes */
sem_t *enc_mutex;       /* Mutex to access encryption keys */
sem_t *node_mutex;      /* route table mutex */

/* Various mqueue descriptors*/
mqd_t parent_write_fd, sender_read_fd, sender_write_fd; 
mqd_t router_read_fd, router_write_fd, receiver_read_fd;

/* Buffer to hold sender , receiver and router mqueue names */
char sender_mqueue[100];
char receiver_mqueue[100];
char router_mqueue[100];

int childpid[MAX_NODES]={0}; /* Array of child pids */
int j;                       /* Child pid array iterator */

char *manets_log = (char *)NULL;

/* Encryption and decryption execution time calculators */
struct timeval start_time, end_time;
int total_usecs, sum_total_usecs;
unsigned int enc_sum=0, enc_avg=0, enc_count=0, enc_msg_split_id =0, enc_avg_sum=0;
unsigned int dec_sum=0, dec_avg=0, dec_count=0;


/* Function prototypes */
/* Various node function */
/***************************************************************************************
  Operation performed by the function start_nodes()
    *Creates the shared memory for symmetric keys and route table.
    *Intilaizes the symmetric key and route table shared memory.
    *Starts up all the nodes, the sender, routers and the receiver.
    *Calls the signal handler when the user uses the command ctrl+c
***************************************************************************************/
void start_nodes();

/***************************************************************************************
  Operation performed by the function sender()
    *Calls the signal handler when the sender receives the SIGINT signal.
    *Calls function add_node() to make an entry of itself in the route table.
    *Creates, opens and reads the message queue to get the user input message.
    *Splits the message into partial messages depending on the 
     SENDER_MSG_SPLIT_COUNT. 
    *Encrypts the partial messages by calling function encrypt().
    *Gets the neighbouring nodes message queues by calling function 
     get_next_node().
    *Opens up the neighbouring nodes message queues and sends the 
     partial messages to these nodes.
***************************************************************************************/
void sender();

/***************************************************************************************
  Operation performed by the function router()
    *Calls the signal handler when the router receives the SIGINT signal.
    *Calls function add_node() to make an entry of itself in the route table.
    *Creates, opens and reads the message queue to get the message 
     sent by the sender.
    *Gets the neighbouring nodes message queues by calling function 
     get_next_node().
    *Opens up the neighbouring nodes message queues and sends the messages 
     to these nodes.
***************************************************************************************/
void router();

/***************************************************************************************
  Operation performed by the function receiver()
    *Calls the signal handler when the receiver receives the SIGINT signal.
    *Calls function add_node() to make an entry of itself in the route table.
    *Creates, opens and reads the message queue to get message sent 
     by the nodes.
    *Decrypts the partial messages by calling functions 
     buff_to_mpz_array and decrypt().
***************************************************************************************/
void receiver();


/* Check system errors */
void errfunction (char error_loc[]             /* Error Location */); 


/* Signal handlers for various nodes */

/***************************************************************************************
  Operation performed by the function parenthandler()
    *Sends the signal SIGINT to all the nodes.
    *Unlinks the symmetric key and route table shared memory.
***************************************************************************************/
void parenthandler(int signum   /* Signal number */);

/***************************************************************************************
  Operation performed by the function senderhandler()
    *Closes the file descriptors opened by the sender process.
    *Unlinks the message queue created by the sender process.
***************************************************************************************/
void senderhandler(int signum   /* Signal number */);

/***************************************************************************************
  Operation performed by the function receiverhandler()
    *Closes the file descriptor opened by the receiver process.
    *Unlinks the message queue created by the receiver process.
***************************************************************************************/
void receiverhandler(int signum /* Signal number */);

/***************************************************************************************
  Operation performed by the function routerhandler()
    *Closes the file descriptor opened by the router process.
    *Unlinks the message queue created by the router process.
***************************************************************************************/
void routerhandler(int signum   /* Signal number */);


/***************************************************************************************
  Operation performed by the function parent_to_sender_msg()
    *Gets the sender message queue by calling function get_sender_mqueue().
    *Calls appropriate functions depending on the user input.
    *Opens up the sender message queue to send the user input message.
***************************************************************************************/
void parent_to_sender_msg();


/***************************************************************************************
  Operation performed by the function create_shared_memory()
    *Creates the symmetric key shared memory by calling function shm_open();
    *Creates the route table shared memory by calling function shm_open();
***************************************************************************************/
void create_shared_memory();


/***************************************************************************************
  Operation performed by the function init_enc_keys()
    *Opens the symmetric key shared memory by calling function shm_open().
    *Sets the values of the symmetric keys to a default value.
***************************************************************************************/
void init_enc_keys();

/***************************************************************************************
  Operation performed by the function init_route_table()
    *Opens the symmetric key shared memory by calling function shm_open().
    *Initializes the  route table counter values.
***************************************************************************************/
void init_route_table();

/***************************************************************************************
  Operation performed by the function set_encryption_scheme()
    *Opens the symmetric key shared memory by calling 
     function get_symmetric_key().
    *Sets the encryption scheme in the shared memory.
***************************************************************************************/
void set_encryption_scheme(enc_scheme encryption_scheme /* Encryption scheme name */);

/***************************************************************************************
  Operation performed by the function get_encryption_scheme()
    *Opens the symmetric key shared memory by calling function 
     get_symmetric_key().
    *Returns the encryption scheme from the shared memory.
***************************************************************************************/
enc_scheme get_encryption_scheme();


/* Various encryption functions */
/***************************************************************************************
  Operation performed by the function encrypt()
    *Calls the appropriate encryption function depending upon the value of 
     the encryption scheme.
***************************************************************************************/
void encrypt(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id  */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
);

/***************************************************************************************
  Operation performed by the function df_add_mul_Ph_encryption()
    *Encrypts the message using Domingo-Ferrer's additive and multiplicative 
     privacy homomorphism.
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void df_add_mul_Ph_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function df_new_Ph_encryption()
    *Encrypts the message using Domingo-Ferrer's new privacy homomorphism. 
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void df_new_Ph_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function df_field_encryption()
    *Encrypts the message using Domingo-Ferrer's field privacy homomorphism.
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void df_field_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function mmh_encryption()
    *Encrypts the message using mixed multiplicative homomorphism. 
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void mmh_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);


/* Various decryption functions */
/***************************************************************************************
  Operation performed by the function decrypt()
    *Calls the appropriate decryption function depending upon the value of 
     the encryption scheme.
***************************************************************************************/
void decrypt(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys*/
);

/***************************************************************************************
  Operation performed by the function df_add_mul_Ph_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is then decrypted by using Domingo-Ferrer's 
     additive and multiplicative privacy homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void df_add_mul_Ph_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

/***************************************************************************************
  Operation performed by the function df_new_Ph_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is then decrypted by using Domingo-Ferrer's 
     new privacy homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void df_new_Ph_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

/***************************************************************************************
  Operation performed by the function df_field_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is decrypted by using Domingo-Ferrer's field 
     privacy homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void df_field_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

/***************************************************************************************
  Operation performed by the function mmh_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is decrypted by using mixed multiplicative homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void mmh_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);


/* Various functions to convert partial message from buffer to mpz array */

/***************************************************************************************
  Operation performed by the function buff_to_mpz_array()
    *Calls the appropriate buff_to_mpz_array decryption function depending upon 
     the value of the encryption scheme.
***************************************************************************************/
void buff_to_mpz_array(
  char receivebuff[]                   /* Partail message in buffer */
  , mpz_t encryption_msg_split_array[] /* Partial message in mpz array */
  , int encryption_split_number        /* Symmetric key */
  , enc_scheme encryption_scheme       /* Encryption scheme */
);

/***************************************************************************************
  Operation performed by the function df_add_mul_buff_to_mpz_array()
    *Places the encrypted message value in the char buffer to the mpz array.
***************************************************************************************/
void df_add_mul_buff_to_mpz_array(
  char receivebuff[]                   /* Partail message in buffer */
  , mpz_t encryption_msg_split_array[] /* Partial message in mpz array */
  , int encryption_split_number        /* Symmetric key */
);

/***************************************************************************************
  Operation performed by the function df_new_buff_to_mpz_array()
    *Places the encrypted message value in the char buffer to the mpz array.
***************************************************************************************/
void df_new_buff_to_mpz_array(
  char receivebuff[]                   /* Partail message in buffer */
  , mpz_t encryption_msg_split_array[] /* Partial message in mpz array*/
  , int encryption_split_number        /* Symmetric key */
);

/***************************************************************************************
  Operation performed by the function df_field_mmh_buff_to_mpz_array()
    *Places the encrypted message value in the char buffer to the mpz array.
***************************************************************************************/
void df_field_mmh_buff_to_mpz_array(
  char receivebuff[]                   /* Partail message in buffer */
  , mpz_t encryption_msg_split_array[] /* Partial message in mpz array*/
);


/* Set various encryption keys in the shared memory */
/***************************************************************************************
  Operation performed by the function set_df_add_mul_ph_keys()
    *Takes in the Domingo_Ferrer's additive multiplicative privacy homomorphism
     symmetric keys as user input.
    *Sets the user input symmetric keys into the symmetric key shared memory. 
***************************************************************************************/
void set_df_add_mul_ph_keys();

/***************************************************************************************
  Operation performed by the function set_df_new_ph_keys()
    *Takes in the Domingo_Ferrer's new  privacy homomorphism symmetric keys 
     as user input.
    *Sets the user input symmetric keys into the symmetric key shared memory. 
***************************************************************************************/
void set_df_new_ph_keys();

/***************************************************************************************
  Operation performed by the function set_df_field_mmh_ph_keys()
    *Takes in the Domingo_Ferrer's field privacy homomorphism and 
     mixed multiplicative homomorphism symmetric keys as user input.
    *Sets the user input symmetric keys into the symmetric key shared memory. 
***************************************************************************************/
void set_df_field_mmh_ph_keys();

/***************************************************************************************
  Operation performed by the function set_df_mul_ph_rd_keys()
    *Chooses the appropriate Domingo_Ferrer's additive multiplicative 
     privacy homomorphism symmetric key as user input depending upon 
     the parameter value.
    *Sets the user input symmetric key into the symmetric key shared memory. 
***************************************************************************************/
void set_df_mul_ph_rd_keys(
  mpz_t value /* Choosing the appropriate symmetric key to be set in shared memory */
);

/***************************************************************************************
  Operation performed by the function set_df_new_ph_rp_p_d_keys()
    *Chooses the appropriate Domingo_Ferrer's new privacy homomorphism 
     symmetric key as user input depending upon the parameter value.
    *Sets the user input symmetric key into the symmetric key shared memory. 
***************************************************************************************/
void set_df_new_ph_rp_p_d_keys(
  mpz_t value /* Choosing the appropriate symmetric key to be set in shared memory */
);

/***************************************************************************************
  Operation performed by the function set_df_field_mmh_ph_p_r_keys()
    *Chooses the appropriate Domingo_Ferrer's field privacy homomorphism 
     and mixed multiplicative privacy homomorphism symmetric key as user input 
     depending upon the parameter value.
    *Sets the user input symmetric key into the symmetric key shared memory. 
***************************************************************************************/
void set_df_field_mmh_ph_p_r_keys(
  mpz_t value /* Choosing the appropriate symmetric key to be set in shared memory */
);


/* Get various encryption keys from the shared memory */
void get_various_encryption_keys(
  mpz_symmetric_key *mpz_keys_ptr /* Structure with the symmetric keys */
);


/* Sets up the message queue, path, level, group and status of the new node */
void add_node(
  int pid /* pid of the node */
);


/* Get the message queues and group id of the neighbouring nodes */
void get_next_node(
  int pid                     /* pid of the node */    
  , char msg_queue[][100]     /* Message queues of the neighbouring node */
  , int neighbouring_groups[] /* group id of the neighbouring nodes */
  , int *count                /* Number of neighbouring nodes */
);


/* Function returning sender mqueue */
void get_sender_mqueue(
  char msg_queue[][100] /* The sender mqueue */
);



/* Generate the seed value for random function */
long get_seed();
 

/* Get the pointer to the shared memory of the symmetric key */ 
symmetric_key *get_symmetric_key();


/* Get the pointer to the shared memory of the route table */ 
route_table *get_route_table();

/* Set the path for either the encryption or decryption log file*/
void set_manets_log(char *);

